<?php

namespace Teknisa\FIN\Modules\ContaPagar\Services;

use Teknisa\FIN\Modules\ContaPagar\Services\Retorno\RetornoFactory;
use Doctrine\ORM\EntityManager;
use Teknisa\Libs\Util\DateUtil;
use Teknisa\Libs\Util\Environment;
use Teknisa\Model\Financeiro\Entities\Cnabpag;
use Teknisa\Model\Financeiro\Entities\Boletpagar;
use Teknisa\Model\Financeiro\Entities\Errcnabbolp;
use Teknisa\Model\Financeiro\Entities\GeralPag;
use Teknisa\FIN\Modules\ContaPagar\Factory\RetornoPagarFactory;
use Teknisa\Libs\Util\NewCode;

class RetornoPagarService {
    CONST TIPO_CNAB_REMESSA = 0;
    CONST TIPO_CNAB_RETORNO = 1;

    CONST OPERACAO_INCLUSAO = 'I';
    CONST OPERACAO_EXCLUSAO = 'E';
    CONST OPERACAO_BLOQUEIO = 'B';

    CONST SITUAACAO_NAO_ENVIADO = '00';
    CONST SITUACAO_ENVIADO_COM_ERRO = '01';
    CONST SITUACAO_ENVIADO_SEM_RETORNO = '10';
    CONST SITUACAO_ENVIADO_RETORNO_OK = '20';
    CONST SITUACAO_BAIXADO_VIA_CNAB = '90';

    const CONTADOR_BAIXA = 'BAIXAPAGA';
    const CONTADOR_CNAB_PAGAMENTO = 'CNABPAG';

    CONST STATUS_PENDENTE_CNAB = 'C';
    CONST STATUS_PAGAMENTO_PLANEJADO = 'P';

    protected $entityManager;
    protected $factory;
    protected $environment;
    protected $novoCodigo;
    protected $baixaInserida;

    public function __construct(EntityManager $entityManager, Environment $environment, RetornoPagarFactory $factory, NewCode $novoCodigo)
    {
        $this->entityManager = $entityManager;
        $this->environment = $environment;
        $this->factory = $factory;
        $this->novoCodigo = $novoCodigo;
    }

    public function validaLayoutBanco($cdbanco, $cdempresa, $nrorg){

        $emprbanc = $this->factory->getEmprbanc($cdbanco, $cdempresa, $nrorg);
        if (!$emprbanc){
            throw new \Exception('Cnab não parametrizada para o banco informado.');
        }
        return $emprbanc->getIdpadrcnabpag();
    }

    public function carregaRetorno($row){
        $nrorg = isset($row['NRORG'])? $row['NRORG'] : $this->environment->getNrOrgTrab();
        $valorCodificado = $row['CNABRETORNO'];
        $valorDecodificado = $this->decodeFile($valorCodificado);
        //$valorDecodificado conteudo do arquivo
        $rowsRetorno = $this->getStringInlines($valorDecodificado);
        $layout = $this->validaLayoutBanco($row['CDBANCO'], $row['CDEMPRESA'], $nrorg);
        $leitorCnab = RetornoFactory::getBancoCnab($row['CDEMPRESA'], $row['CDBANCCNAB'], $rowsRetorno, $layout);
        $leitorCnab->leArquivo($nrorg);
        $leitorCnab->validaBoletos($nrorg);
        $dataSet = array();
        $dataSet[] = $leitorCnab->getRetornos();
        $dataSet[] = $leitorCnab->getErros();
        $dataSet[] = [$leitorCnab->getStatusArquivo()];
        return $dataSet;
    }

    public function decodeFile($encodedFile) {
        $file = base64_decode($encodedFile);
        return $file;
    }

    public function getStringInlines($file) {
        $file = preg_split('/\R/', $file);
        return $file;
    }
    
    /*
     * Insere Retorno Pagamento
     */
    public function insereRetorno(
        $boletos, $ocorrencias, $dsarqcnabpag, $cdempresa, $cdbanco,
        $nrorg = false, $cdfilial = false, $cdoperador = false, $importacao = false
    ){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try{
            $this->insereRetornoPagamento(
                $boletos, $ocorrencias, $dsarqcnabpag, $cdempresa, $cdbanco,
                $nrorg, $cdfilial, $cdoperador, $importacao
            );
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollback();
            throw $e;
        }
    }

    /*
     * Insere Retorno Pagamento
     */
    public function insereRetornoPagamento(
        $boletos, $ocorrencias, $dsarqcnabpag, $cdempresa, $cdbanco,
        $nrorg = false, $cdfilial = false, $cdoperador = false, $importacao = false
    ){
        try{
            
            // Preenche Váriaveis Iniciais
            $nrorg      = !$nrorg      ? $this->environment->getNrOrgTrab()  : $nrorg;
            $cdfilial   = !$cdfilial   ? $this->environment->getCdFilial()   : $cdfilial;
            $cdoperador = !$cdoperador ? $this->environment->getCdOperador() : $cdoperador;
            
            // Gera Novo Código
            $prxcnabpag = $this->novoCodigo->getCodeByTableName(self::CONTADOR_CNAB_PAGAMENTO.$cdempresa, $nrorg);
            
            // Insere Retorno Pagamento
            $cnabpag = $this->factory->insereCnabPag($nrorg, $cdempresa, $prxcnabpag, self::TIPO_CNAB_RETORNO, $cdoperador, $cdbanco, $dsarqcnabpag);
            $this->entityManager->persist($cnabpag);
            $this->entityManager->flush();

            // Prepara Títulos p/ Inclusão no Retorno Pagamento
            $boletosBaixa = $this->preparaBoletoInclusao($boletos, $importacao);
            foreach ($boletosBaixa as $data=>$boletosData) {
                // Percorre os Tipos de Baixa
                foreach ($boletosData as $tpbaixa=>$boletosTpBaixa) {
                    
                    // Gera Novo Código Baixa
                    $prxbaixa = $this->novoCodigo->getCodeByTableName(self::CONTADOR_BAIXA.$cdfilial, $nrorg);

                    // Variavel de controle para inserir uma baixa por Tipo de baixa
                    $this->baixaInserida = false;

                    foreach ($boletosTpBaixa as $key=>$boleto) {
                        $valortotal = 0;
                        if($boleto['IDSITUCNABP'] == self::SITUACAO_BAIXADO_VIA_CNAB){
                            if(round($boleto['VRPAGO'],2) != round($boleto['VRFINALBPAG'],2)){
                                $valortotal+= ($boleto['VRFINALBPAG'] +
                                               $boleto['VRACRESCIMO'] -
                                               $boleto['VRDESCONTO']);
                            }else{
                                $valortotal+= ($boleto['VRFINALBPAG']);
                            }
                            $baixapaga = $this->validaBaixaBoleto($data, $cdempresa, $cdfilial, $tpbaixa,
                                $boleto['CDREDUCONT'], $cdoperador, $nrorg, $prxbaixa, $valortotal, $prxcnabpag);
                            $this->updateSituacaoBoleto($boleto, $valortotal, $baixapaga->getCdfilial(), $prxbaixa, $prxcnabpag, $nrorg);

                            /** @var GeralPag $geralpag */
                            $geralpag = $this->factory->getGeralPag($nrorg);

                            // Insere Ajuste de Acréscimo
                            if ($geralpag->getCdajacrepag() && $boleto['VRACRESCIMO'] > 0){
                                $ajusteacres = $this->factory->inserirAjusteBoleto($nrorg, $boleto['CDFILIAL'], $boleto['NRLANBOLPAG'], $geralpag->getCdajacrepag(), $boleto['VRACRESCIMO']);
                                $this->entityManager->persist($ajusteacres);
                            }

                            // Insere Ajuste de Desconto
                            if($geralpag->getCdajdescpag() && ($boleto['VRDESCONTO'] > 0)) {
                                $ajustedesc = $this->factory->inserirAjusteBoleto($nrorg, $boleto['CDFILIAL'], $boleto['NRLANBOLPAG'], $geralpag->getCdajdescpag(), $boleto['VRDESCONTO']);
                                $this->entityManager->persist($ajustedesc);
                            }

                            $this->entityManager->flush();

                            if($ocorrencias){
                                $this->insereOcorrenciasBoleto($boleto, $ocorrencias, $cdbanco, $nrorg);
                            }
                        }else{
                            $boletpagar = $this->factory->getBoletoPagar($boleto['CDFILIAL'], $boleto['NRLANBOLPAG'], $nrorg);
                            $boletpagar = $this->factory->alteraSituacaoCnabPag($boletpagar, $prxcnabpag, $boleto['IDSITUCNABP']);
                            $this->entityManager->persist($boletpagar);
                        }
                    }
                }
            }
            if(empty($boletosBaixa) && $importacao){
                var_dump('Não há boletos aptos a serem importados.');
                return false;
            } else {
                $this->entityManager->flush();
                return true;
            }
        } catch (\Exception $e) {
            throw $e;
        }
    }

    public function preparaBoletoInclusao($boletos, $importacao = false){
        $boletosporData = array();
        foreach ($boletos as $boleto) {
            if ($boleto['SITUACAOBOLETO']) {
                //Agrupa os boletos por Data de Pagamento
                $boletosporData[$boleto['DTPAGAMENTO']][] = $boleto;
            }
        }
        if((!$boletosporData) && $importacao === false){
            throw new \Exception('Não há boletos aptos a serem importados.');
        } else if((!$boletosporData) && $importacao){
            return [];
        }
        //Agrupa Boletos por Tipo de Baixa
        $boletosBaixa = array();
        foreach ($boletosporData as $key => $boletosData) {
            foreach ($boletosData as $boleto) {
                $boletosBaixa[$boleto['DTPAGAMENTO']][$boleto['CDTPBAIXAPAG']][] = $boleto;
            }
        }
        return $boletosBaixa;
    }

    public function validaBaixaBoleto($data, $cdempresa, $cdfilial, $tpbaixa,
        $cdreducont, $operador, $nrorg, $prxbaixa, $valortotal, $prxcnabpag){
        if(!$this->baixaInserida){
            $dtpagamento = str_replace('/', '', $data);
            $dtpagamento = substr($dtpagamento, 0, 2).'/'.substr($dtpagamento, 2, 2).'/'.substr($dtpagamento, 4, 4);
            $dtpagamento = self::formataData($dtpagamento);

            $baixapaga = $this->factory->insereBaixa($cdempresa, $cdfilial, $tpbaixa, $dtpagamento,
                $cdreducont, $operador, $nrorg, $prxbaixa, $valortotal, $prxcnabpag, self::STATUS_PENDENTE_CNAB);
            $this->baixaInserida = true;
        }else{
            $baixapaga = $this->factory->getBaixaPaga($cdfilial, $prxbaixa, $nrorg);
            $baixapaga = $this->factory->insereValorBaixa($baixapaga, $valortotal);
        }
        $this->entityManager->persist($baixapaga);
        $this->entityManager->flush();
        return $baixapaga;
    }

    public function updateSituacaoBoleto($boleto, $valortotal, $cdfilibaipag, $prxbaixa, $prxcnabpag, $nrorg){
        $boletpagar = $this->factory->getBoletoPagar($boleto['CDFILIAL'], $boleto['NRLANBOLPAG'], $nrorg);
        $boletpagar = $this->factory->alteraBoleto($boletpagar, $valortotal, $cdfilibaipag, $prxbaixa,
            $prxcnabpag, $boleto['IDSITUCNABP'], self::STATUS_PAGAMENTO_PLANEJADO);
        $this->entityManager->persist($boletpagar);
        $this->entityManager->flush();
    }


    public function insereOcorrenciasBoleto($boleto, $ocorrencias, $cdbanco, $nrorg){
        //Exclui erros gravados anteriormente
        $erroscnab = $this->factory->getErrosCnab($boleto['NRLANBOLPAG'],
                                      $boleto['CDFILIAL'],
                                      $nrorg);
        foreach ($erroscnab as $errocnab) {
            $this->entityManager->remove($errocnab);
        }
        $this->entityManager->flush();
        foreach ($ocorrencias as $ocorrencia) {
            if($boleto['CDLINHA'] == $ocorrencia['CDLINHA']){
                $erro = $this->factory->insereOcorencia($boleto['CDFILIAL'],
                                               $boleto['NRLANBOLPAG'],
                                               $cdbanco,
                                               $ocorrencia['CDERRO'],
                                               $nrorg
                                               );
                $this->entityManager->persist($erro);
            }
        }
    }

    /**
     * Cancela Titulos do Retorno
     * @param $retorno
     * @param $boletos
     * @throws \Exception
     */
    public function cancelaTitulosRetorno($retorno, $boletos){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try{
            foreach($boletos as $boleto)
                self::cancelaTituloRetorno($retorno, $boleto);

            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollback();
            throw $e;
        }
    }

    /**
     * Cancela Titulo do Retorno
     * @param $retorno
     * @param $boleto
     * @throws \Exception
     */
    public function cancelaTituloRetorno($retorno, $boleto){
        try{
            // Busca Cnab Pagar
            /** @var Cnabpag $cnabpag */
            $cnabpag = $this->factory->getCnabpag(
                $retorno['CDEMPRESA'], $retorno['NRCNABPAG'], $retorno['NRORG']
            );

            if($cnabpag) {

                // Busca Boleto Cobrança
                /** @var Boletpagar $boletpagar */
                $boletpagar = $this->factory->getBoletoRetorno(
                    $boleto['CDFILIAL'], $boleto['NRLANBOLPAG'], $boleto['NRORG'], $cnabpag->getNrcnabpag()
                );

                if ($boletpagar) {
                    // Busca Ocorrências do Boleto Cobrança
                    /** @var Errcnabbolp $errcnabbolrs */
                    $errcnabbolps = $this->factory->getErrosCnab(
                        $boletpagar->getNrlanbolpag(), $boletpagar->getCdfilial(), $boletpagar->getNrorg()
                    );

                    // Exclui Ocorrências Existêntes do Boleto Cobrança
                    if ($errcnabbolps)
                        foreach ($errcnabbolps as $errcnabbolp)
                            $this->entityManager->remove($errcnabbolp);

                    // Atualiza Boleto Cobrança
                    $boletpagar->setNrcnabpagret('');
                    $boletpagar->setIdsitucnabp('00');
                    $this->entityManager->persist($boletpagar);

                    // Atualiza Retorno Cobrança
                    $cnabpag->setVrtotalcnabpag($cnabpag->getVrtotalcnabpag() - $boletpagar->getVrfinalbpag());
                    $cnabpag->setQtbolcnabpag($cnabpag->getQtbolcnabpag() - 1);
                    $this->entityManager->persist($cnabpag);
                }
            }

            $this->entityManager->flush();
        } catch (\Exception $e) {
            throw $e;
        }
    }

    public static function formataData(
        $data
    ){
        if (!$data)
            return null;

        if (strlen($data) == 19)
            $data = DateUtil::truncateString($data);
        $data = DateUtil::getDataDeString($data);
        $data = DateUtil::truncateData($data);

        return $data;
    }

}